/* -*- mode: C; c-file-style: "stroustrup"; indent-tabs-mode: nil; -*- 
 */ 

/*
 * This file is part of hildon-navigator
 *
 * Copyright (C) 2005 Nokia Corporation.
 *
 * Contact: Karoliina Salminen <karoliina.t.salminen@nokia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#ifndef HILDON_NAVIGATOR_WM_H
#define HILDON_NAVIGATOR_WM_H

#include <X11/Xlib.h>
#include <X11/Xatom.h>

#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkevents.h>

#include <gtk/gtk.h>

#include <hildon-navigator.h>

#include <libmb/mbdotdesktop.h>

#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>

#define DESKTOP_BIN_FIELD   "Exec"
#define DESKTOP_LAUNCH_FIELD    "X-Osso-Service"
#define DESKTOP_VISIBLE_FIELD "Name"



#define DUMMY_STRING " \0"

#define KILL_APPS_METHOD "kill_app"

/* The window types that we care about */

enum {
    OTHER = 0,
    UNKNOWN,
    NORMAL_WINDOW,
    MODAL_DIALOG,
    DESKTOP,
    MENU
};



/* Callback definitions*/

/** Function callback to be called when a new window/view
 *  is created.
 *  @param icon_name The name of icon for this menu entry
 *  @param app_name The name of application for this menu entry
 *  @param view_name The title for this window/view
 *  @param data The callback data
 *  @return GtkWidget pointer to this menu item
 */

typedef GtkWidget *(wm_new_window_cb)(const gchar *icon_name,
                                      const gchar *app_name,
                                      const gchar *view_name,
                                      gpointer data);

/** Function callback to be called when a window is removed.
 *
 *  @param menuitem Pointer to the menuitem that corresponds with
 *                  the window that was removed
 *  @param data Pointer to callback data
 */

typedef void (wm_removed_window_cb)(GtkWidget *menuitem,
                                     gpointer data);


/** Function callback to be called when a window has been changed.
 *  Mostly used to signal when the name of the view is set,
 *  because this has probably not yet happened when the CreateNotify
 *  event is handled.
 *
 * @param menuitem Pointer to the menuitem that corresponds with
 *                 the window which properties changed
 * @param icon_name The name of icon for this menu entry
 * @param app_name The name of application for this menu entry
 * @param view_name The title for this window/view
 * @param position_change 0 if no change, 1 if item shall migrate to
 *                        first position, 2 if item shall migrate to
 *                        the last position (i.e. iconized)
 *
 * @param killable TRUE if the application can be killed
 * @param dialog_name Contains the title of the dialog window
 * @param data The callback data
 *
 */

typedef void (wm_updated_window_cb)(GtkWidget *menuitem,
                                    const gchar *icon_name,
                                    const gchar *app_name,
                                    const gchar *view_name,
                                    const guint position_change,
                                    const gboolean killable,
                                    const gchar *dialog_name,
                                    gpointer data);


/** Function callback to be called when the desktop has been topped.
 * @ param The callback data
 */

typedef void (wm_desktop_topped_cb)(gpointer data);






/* public functions of the Task Navigator 'window manager' */

/**
 * @param new_window_cb Pointer to a function that the window manager
 *        calls when a new window is created
 * @param removed_window_cb Pointer to a function that the window manager
 *        calls when a window is removed
 * @param updated_win_cb Pointer to a function that the window manager
 *        calls when properties of a window are updated
 * @param desktop_topped_cb Pointer to a function that is called when
 *        the desktop (i.e. Home) is topped.
 * @param cb_data Data to be passed to the callbacks
 * @return TRUE if initialization succeeded, FALSE otherwise.
 */

gboolean init_window_manager(wm_new_window_cb *new_win_cb,
                             wm_removed_window_cb *removed_win_cb,
                             wm_updated_window_cb *updated_win_cb,
                             wm_desktop_topped_cb *desktop_topped_cb,
                             gpointer cb_data);


/**  Send 'top' request for a certain existing window/view
 *
 * @param menuitem The menuitem for the view/window that is to be topped.
 *
 */

void top_view(GtkMenuItem *menuitem);


/**  Send 'top' request for a certain service
 *   @param service_name The name of the service that is to be topped
 */

void top_service(const gchar *service_name);

/**
 * Requests the real window manager to top the desktop
 */

void top_desktop(void);


#endif /* HILDON_NAVIGATOR_WM_H */
